const debug = require('debug')('telegraf:core')
const Telegram = require('./telegram/telegram')
const Composer = require('./composer')
const Context = require('./context')

class Telegraf extends Composer {

  constructor (token, options) {
    super()
    this.options = Object.assign({ retryAfter: 2000 }, options)
    this.token = token
    this.onError = (err) => {
      console.error()
      console.error((err.stack || err.toString()).replace(/^/gm, '  '))
      console.error()
      throw err
    }
    this.context = {}
    this.pollingState = {
      offset: 0,
      started: false
    }
  }

  set token (token) {
    this.options.token = token
    this.telegram = new Telegram(this.options.token, this.options.telegram)
  }

  get token () {
    return this.options.token
  }

  catch (handler) {
    this.onError = handler
    return this
  }

  startPolling (timeout, limit) {
    const isStarted = this.pollingState.started
    this.pollingState = Object.assign(this.pollingState, {
      started: true,
      timeout: timeout || 30,
      limit: limit || 100
    })
    if (!isStarted) {
      this.fetchUpdates()
    }
    return this
  }

  webHookCallback (webHookPath = '/') {
    return (req, res, next) => {
      if (req.method !== 'POST' || req.url !== `${webHookPath}`) {
        if (next && typeof next === 'function') {
          return next()
        }
        res.statusCode = 403
        return res.end()
      }
      let body = ''
      req.on('data', (chunk) => {
        body += chunk.toString()
      })
      req.on('end', () => {
        let update = {}
        try {
          update = JSON.parse(body)
        } catch (error) {
          res.writeHead(415)
          res.end()
          return this.onError(error)
        }
        this.handleUpdate(update, res)
          .then(() => {
            if (!res.finished) {
              res.end()
            }
          })
          .catch((err) => {
            debug('webhook error', err)
            res.writeHead(500)
            res.end()
          })
      })
    }
  }

  startWebHook (webHookPath, tlsOptions, port, host) {
    const callback = this.webHookCallback(webHookPath)
    this.webhookServer = tlsOptions
      ? require('https').createServer(tlsOptions, callback)
      : require('http').createServer(callback)
    this.webhookServer.listen(port, host, () => {
      debug('WebHook listening on port: %s', port)
    })
    return this
  }

  stop () {
    this.pollingState.started = false
    if (this.webhookServer) {
      this.webhookServer.close()
    }
    return this
  }

  handleUpdate (update, webHookResponse) {
    debug('⚡ update', update.update_id)
    const telegram = webHookResponse ? new Telegram(this.token, this.options.telegram, webHookResponse) : this.telegram
    const ctx = new Context(update, telegram, this.options)
    Object.assign(ctx, this.context)
    return this.middleware()(ctx).catch(this.onError)
  }

  fetchUpdates () {
    const pollingState = this.pollingState
    if (!pollingState.started) {
      return
    }
    this.telegram.getUpdates(pollingState.timeout, pollingState.limit, pollingState.offset)
      .catch((err) => {
        console.error('Telegraf: network error', err.message)
        return new Promise((resolve) => setTimeout(resolve, err.retryAfter || this.options.retryAfter, []))
      })
      .then((updates) => {
        if (!Array.isArray(updates)) {
          return []
        }
        const tasks = updates.map((update) => this.handleUpdate(update))
        return Promise.all(tasks).then(() => updates)
      })
      .catch((err) => {
        console.error('Telegraf: polling error', err.message)
        pollingState.offset = 0
        pollingState.started = false
        return []
      })
      .then((updates) => {
        if (updates.length > 0) {
          pollingState.offset = updates[updates.length - 1].update_id + 1
        }
        this.fetchUpdates()
      })
  }
}

module.exports = Telegraf
